@@ -1,5 +1,5 @@
 /*!
- * Copyright (c) 2017-2020 by Contributors
+ * Copyright (c) 2017-2023 by Contributors
  * \file c_api.h
  * \author Hyunsu Cho
  * \brief C API of Treelite, used for interfacing with other languages
@@ -31,11 +31,12 @@ typedef void* AnnotationHandle;
 typedef void* CompilerHandle;
 /*! \brief handle to a polymorphic value type, used in the model builder API */
 typedef void* ValueHandle;
+/*! \brief handle to a configuration of GTIL predictor */
+typedef void* GTILConfigHandle;
 /*! \} */
 
 /*!
- * \defgroup annotator
- * Branch annotator interface
+ * \defgroup annotator Branch annotator interface
  * \{
  */
 /*!
@@ -67,8 +68,7 @@ TREELITE_DLL int TreeliteAnnotationFree(AnnotationHandle handle);
 /*! \} */
 
 /*!
- * \defgroup compiler
- * Compiler interface
+ * \defgroup compiler Compiler interface
  * \{
  */
 /*!
@@ -110,71 +110,132 @@ TREELITE_DLL int TreeliteCompilerFree(CompilerHandle handle);
 /*! \} */
 
 /*!
- * \defgroup model_loader
- * Model loader interface: read trees from the disk
+ * \defgroup model_loader Model loader interface: read trees from the disk
  * \{
  */
 /*!
- * \brief load a model file generated by LightGBM (Microsoft/LightGBM). The
- *        model file must contain a decision tree ensemble.
- * \param filename name of model file
- * \param out loaded model
- * \return 0 for success, -1 for failure
+ * \brief Deprecated. Please use \ref TreeliteLoadLightGBMModelEx instead.
  */
 TREELITE_DLL int TreeliteLoadLightGBMModel(const char* filename, ModelHandle* out);
 /*!
- * \brief load a model file generated by XGBoost (dmlc/xgboost). The model file
- *        must contain a decision tree ensemble.
+ * \brief Load a model file generated by LightGBM (Microsoft/LightGBM). The
+ *        model file must contain a decision tree ensemble.
  * \param filename name of model file
+ * \param config_json null-terminated JSON string consisting key-value pairs; used for configuring
+ *                    the model parser
  * \param out loaded model
  * \return 0 for success, -1 for failure
  */
+TREELITE_DLL int TreeliteLoadLightGBMModelEx(const char* filename,
+                                             const char* config_json,
+                                             ModelHandle* out);
+/*!
+ * \brief Deprecated. Please use \ref TreeliteLoadXGBoostModelEx instead.
+ */
 TREELITE_DLL int TreeliteLoadXGBoostModel(const char* filename, ModelHandle* out);
 /*!
- * \brief load a json model file generated by XGBoost (dmlc/xgboost). The model
- *        file must contain a decision tree ensemble.
+ * \brief Load a model file generated by XGBoost (dmlc/xgboost). The model file
+ *        must contain a decision tree ensemble.
  * \param filename name of model file
+ * \param config_json JSON string consisting key-value pairs; used for configuring the model parser
  * \param out loaded model
  * \return 0 for success, -1 for failure
  */
+TREELITE_DLL int TreeliteLoadXGBoostModelEx(const char* filename,
+                                            const char* config_json,
+                                            ModelHandle* out);
+/*!
+ * \brief Deprecated. Please use \ref TreeliteLoadXGBoostJSONEx instead.
+ */
 TREELITE_DLL int TreeliteLoadXGBoostJSON(const char* filename,
                                          ModelHandle* out);
 /*!
- * \brief load a model stored as JSON stringby XGBoost (dmlc/xgboost). The model
- *        json must contain a decision tree ensemble.
- * \param json_str the string containing the JSON model specification
- * \param length the length of the JSON string
+ * \brief Load a json model file generated by XGBoost (dmlc/xgboost). The model
+ *        file must contain a decision tree ensemble.
+ * \param filename name of model file
+ * \param config_json null-terminated JSON string consisting key-value pairs; used for configuring
+ *                    the model parser
  * \param out loaded model
  * \return 0 for success, -1 for failure
  */
+TREELITE_DLL int TreeliteLoadXGBoostJSONEx(const char* filename,
+                                           const char* config_json,
+                                           ModelHandle* out);
+/*!
+ * \brief Deprecated. Please use \ref TreeliteLoadXGBoostJSONStringEx instead.
+ */
 TREELITE_DLL int TreeliteLoadXGBoostJSONString(const char* json_str,
                                                size_t length,
                                                ModelHandle* out);
 /*!
- * \brief load an XGBoost model from a memory buffer.
- * \param buf memory buffer
- * \param len size of memory buffer
+ * \brief Load a model stored as JSON string by XGBoost (dmlc/xgboost). The model
+ *        json must contain a decision tree ensemble.
+ * \param json_str the string containing the JSON model specification
+ * \param length the length of the JSON string
+ * \param config_json null-terminated JSON string consisting key-value pairs; used for configuring
+ *                    the model parser
  * \param out loaded model
  * \return 0 for success, -1 for failure
  */
+TREELITE_DLL int TreeliteLoadXGBoostJSONStringEx(const char* json_str,
+                                                 size_t length,
+                                                 const char* config_json,
+                                                 ModelHandle* out);
+/*!
+ * \brief Deprecated. Please use \ref TreeliteLoadXGBoostModelFromMemoryBufferEx instead.
+ */
 TREELITE_DLL int TreeliteLoadXGBoostModelFromMemoryBuffer(const void* buf,
                                                           size_t len,
                                                           ModelHandle* out);
-
 /*!
- * \brief Load a LightGBM model from a string. The string should be created with the
- *        model_to_string() method in LightGBM.
- * \param model_str the model string
+ * \brief Load an XGBoost model from a memory buffer.
+ * \param buf memory buffer
+ * \param len size of memory buffer
+ * \param config_json null-terminated JSON string consisting key-value pairs; used for configuring
+ *                    the model parser
  * \param out loaded model
  * \return 0 for success, -1 for failure
  */
+TREELITE_DLL int TreeliteLoadXGBoostModelFromMemoryBufferEx(const void* buf,
+                                                            size_t len,
+                                                            const char* config_json,
+                                                            ModelHandle* out);
+
+/*!
+ * \brief Deprecated. Please use \ref TreeliteLoadLightGBMModelFromStringEx instead.
+ */
 TREELITE_DLL int TreeliteLoadLightGBMModelFromString(const char* model_str,
                                                      ModelHandle* out);
 
+/*!
+ * \brief Load a LightGBM model from a string. The string should be created with the
+ *        model_to_string() method in LightGBM.
+ * \param model_str the model string
+ * \param config_json null-terminated JSON string consisting key-value pairs; used for configuring
+ *                    the model parser
+ * \param out loaded model
+ * \return 0 for success, -1 for failure
+ */
+TREELITE_DLL int TreeliteLoadLightGBMModelFromStringEx(const char* model_str,
+                                                       const char* config_json,
+                                                       ModelHandle* out);
+
+/*!
+ * \brief Construct a new Treelite model from a JSON string.
+ * \param json_str JSON string
+ * \param config_json Configuration to use when parsing the JSON string. Configuration should be
+ *                    a JSON object consisting of key-value pairs.
+ * \param out Constructed model
+ * \return 0 for success, -1 for failure
+ */
+TREELITE_DLL int TreeliteBuildModelFromJSONString(const char* json_str,
+                                                  const char* config_json,
+                                                  ModelHandle* out);
+
 /*!
  * \brief Load a scikit-learn random forest regressor model from a collection of arrays. Refer to
  *        https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html to
- *        learn the mearning of the arrays in detail. Note that this function can also be used to
+ *        learn the meaning of the arrays in detail. Note that this function can also be used to
  *        load an ensemble of extremely randomized trees (sklearn.ensemble.ExtraTreesRegressor).
  * \param n_estimators number of trees in the random forest
  * \param n_features number of features in the training data
@@ -207,7 +268,7 @@ TREELITE_DLL int TreeliteLoadSKLearnRandomForestRegressor(
 /*!
  * \brief Load a scikit-learn isolation forest model from a collection of arrays. Refer to
  *        https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html to
- *        learn the mearning of the arrays in detail.
+ *        learn the meaning of the arrays in detail.
  * \param n_estimators number of trees in the random forest
  * \param n_features number of features in the training data
  * \param node_count node_count[i] stores the number of nodes in the i-th tree
@@ -239,7 +300,7 @@ TREELITE_DLL int TreeliteLoadSKLearnIsolationForest(
 /*!
  * \brief Load a scikit-learn random forest classifier model from a collection of arrays. Refer to
  *        https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html to
- *        learn the mearning of the arrays in detail. Note that this function can also be used to
+ *        learn the meaning of the arrays in detail. Note that this function can also be used to
  *        load an ensemble of extremely randomized trees (sklearn.ensemble.ExtraTreesClassifier).
  * \param n_estimators number of trees in the random forest
  * \param n_features number of features in the training data
@@ -273,8 +334,8 @@ TREELITE_DLL int TreeliteLoadSKLearnRandomForestClassifier(
 /*!
  * \brief Load a scikit-learn gradient boosting regressor model from a collection of arrays. Refer
  *        to https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html to
- *        learn the mearning of the arrays in detail.
- * \param n_estimators number of trees in the random forest
+ *        learn the meaning of the arrays in detail.
+ * \param n_iter Number of boosting iterations
  * \param n_features number of features in the training data
  * \param node_count node_count[i] stores the number of nodes in the i-th tree
  * \param children_left children_left[i][k] stores the ID of the left child node of node k of the
@@ -297,7 +358,7 @@ TREELITE_DLL int TreeliteLoadSKLearnRandomForestClassifier(
  * \return 0 for success, -1 for failure
  */
 TREELITE_DLL int TreeliteLoadSKLearnGradientBoostingRegressor(
-    int n_estimators, int n_features, const int64_t* node_count, const int64_t** children_left,
+    int n_iter, int n_features, const int64_t* node_count, const int64_t** children_left,
     const int64_t** children_right, const int64_t** feature, const double** threshold,
     const double** value, const int64_t** n_node_samples, const double** weighted_n_node_samples,
     const double** impurity, ModelHandle* out);
@@ -305,8 +366,8 @@ TREELITE_DLL int TreeliteLoadSKLearnGradientBoostingRegressor(
 /*!
  * \brief Load a scikit-learn gradient boosting classifier model from a collection of arrays. Refer
  *        to https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html to
- *        learn the mearning of the arrays in detail.
- * \param n_estimators number of trees in the random forest
+ *        learn the meaning of the arrays in detail.
+ * \param n_iter Number of boosting iterations
  * \param n_features number of features in the training data
  * \param n_classes number of classes in the target variable
  * \param node_count node_count[i] stores the number of nodes in the i-th tree
@@ -330,11 +391,88 @@ TREELITE_DLL int TreeliteLoadSKLearnGradientBoostingRegressor(
  * \return 0 for success, -1 for failure
  */
 TREELITE_DLL int TreeliteLoadSKLearnGradientBoostingClassifier(
-    int n_estimators, int n_features, int n_classes, const int64_t* node_count,
+    int n_iter, int n_features, int n_classes, const int64_t* node_count,
     const int64_t** children_left, const int64_t** children_right, const int64_t** feature,
     const double** threshold, const double** value, const int64_t** n_node_samples,
     const double** weighted_n_node_samples, const double** impurity, ModelHandle* out);
 
+/*!
+ * \brief Load a scikit-learn HistGradientBoostingRegressor model from a collection of arrays.
+ *        Unlike other scikit-learn models, this model class natively handles missing values like
+ *        XGBooost does.
+ * \param n_iter Number of boosting iterations
+ * \param n_features Number of features in the training data
+ * \param node_count node_count[i] stores the number of nodes in the i-th tree
+ * \param children_left children_left[i][k] stores the ID of the left child node of node k of the
+ *                      i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param children_right children_right[i][k] stores the ID of the right child node of node k of the
+ *                       i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param feature feature[i][k] stores the ID of the feature used in the binary tree split at node k
+ *                of the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param threshold threshold[i][k] stores the threshold used in the binary tree split at node k of
+ *                  the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param default_left default_left[i][k] indicates how the missing value should be handled at
+ *                     node k of the i-th tree. This flag is defined if node k is an internal
+ *                     (non-leaf) node. If True, the missing value will be associated with the left
+ *                     child; if False, the missing value will be associated with the right child.
+ * \param value value[i][k] stores the leaf output of node k of the i-th tree. This is only defined
+ *              if node k is a leaf node.
+ * \param n_node_samples n_node_samples[i][k] stores the number of data samples associated with
+ *                       node k of the i-th tree.
+ * \param gain gain[i][k] stores the gain (reduction of the loss function) associate with node k of
+ *             the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param baseline_prediction Baseline predictions for outputs. At prediction, margin scores will be
+ *                            adjusted by this amount before applying the post-processing (link)
+ *                            function. Required shape: (1,)
+ * \param out pointer to store the loaded model
+ * \return 0 for success, -1 for failure
+ */
+TREELITE_DLL int TreeliteLoadSKLearnHistGradientBoostingRegressor(
+    int n_iter, int n_features, const int64_t* node_count, const int64_t** children_left,
+    const int64_t** children_right, const int64_t** feature, const double** threshold,
+    const int8_t** default_left, const double** value, const int64_t** n_node_samples,
+    const double** gain, const double* baseline_prediction, ModelHandle* out);
+
+/*!
+ * \brief Load a scikit-learn HistGradientBoostingClassifier model from a collection of arrays.
+ *        Unlike other scikit-learn models, this model class natively handles missing values like
+ *        XGBooost does.
+ * \param n_iter Number of boosting iterations
+ * \param n_features Number of features in the training data
+ * \param n_classes Number of classes in the target variable
+ * \param node_count node_count[i] stores the number of nodes in the i-th tree
+ * \param children_left children_left[i][k] stores the ID of the left child node of node k of the
+ *                      i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param children_right children_right[i][k] stores the ID of the right child node of node k of the
+ *                       i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param feature feature[i][k] stores the ID of the feature used in the binary tree split at node k
+ *                of the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param threshold threshold[i][k] stores the threshold used in the binary tree split at node k of
+ *                  the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param default_left default_left[i][k] indicates how the missing value should be handled at
+ *                     node k of the i-th tree. This flag is defined if node k is an internal
+ *                     (non-leaf) node. If True, the missing value will be associated with the left
+ *                     child; if False, the missing value will be associated with the right child.
+ * \param value value[i][k] stores the leaf output of node k of the i-th tree. This is only defined
+ *              if node k is a leaf node.
+ * \param n_node_samples n_node_samples[i][k] stores the number of data samples associated with
+ *                       node k of the i-th tree.
+ * \param gain gain[i][k] stores the gain (reduction of the loss function) associate with node k of
+ *             the i-th tree. This is only defined if node k is an internal (non-leaf) node.
+ * \param baseline_prediction Baseline predictions for outputs. At prediction, margin scores will be
+ *                            adjusted by this amount before applying the post-processing (link)
+ *                            function. Required shape: (1,) for binary classification;
+ *                            (n_classes,) for multi-class classification
+ * \param out pointer to store the loaded model
+ * \return 0 for success, -1 for failure
+ */
+TREELITE_DLL int TreeliteLoadSKLearnHistGradientBoostingClassifier(
+    int n_iter, int n_features, int n_classes, const int64_t* node_count,
+    const int64_t** children_left, const int64_t** children_right, const int64_t** feature,
+    const double** threshold, const int8_t** default_left, const double** value,
+    const int64_t** n_node_samples, const double** gain, const double* baseline_prediction,
+    ModelHandle* out);
+
 /*!
  * \brief Query the number of trees in the model
  * \param handle model to query
@@ -379,7 +517,7 @@ TREELITE_DLL int TreeliteSerializeModel(const char* filename, ModelHandle handle
  * \brief Deserialize (load) a model object from disk
  * \param filename name of the file from which to deserialize the model. The file should be created
  *                 by a call to TreeliteSerializeModel().
- * \param handle handle to the model object
+ * \param out Handle to the model object
  * \return 0 for success, -1 for failure
  */
 TREELITE_DLL int TreeliteDeserializeModel(const char* filename, ModelHandle* out);
@@ -413,46 +551,85 @@ TREELITE_DLL int TreeliteFreeModel(ModelHandle handle);
 /*! \} */
 
 /*!
- * \defgroup gtil
- * General Tree Inference Library (GTIL), providing a reference implementation for
+ * \defgroup gtil General Tree Inference Library (GTIL), providing a reference implementation for
  * predicting with decision trees. GTIL is useful in cases it is infeasible to build the
  * tree models as native shared libs.
  * \{
  */
 
+/*!
+ * \brief Load a configuration for GTIL predictor from a JSON string.
+ * \param config_json a JSON string with the following fields:
+ *   - "nthread" (optional): Number of threads used for initializing DMatrix.
+ *                           Set <= 0 to use all CPU cores.
+ *   - "predict_type" (required): Must be one of the following.
+ *     - "default": Sum over trees and apply post-processing
+ *     - "raw": Sum over trees, but don't apply post-processing; get raw margin scores instead.
+ *     - "leaf_id": Output one (integer) leaf ID per tree.
+ *     - "score_per_tree": Output one or more margin scores per tree.
+ * \param out Parsed configuration
+ * \return 0 for success; -1 for failure
+ */
+TREELITE_DLL int TreeliteGTILParseConfig(const char* config_json, GTILConfigHandle* out);
+
+/*!
+ * \brief Delete a GTIL configuration from memory
+ * \param handle Handle to the GTIL configuration to be deleted
+ * \return 0 for success; -1 for failure
+ */
+TREELITE_DLL int TreeliteGTILDeleteConfig(GTILConfigHandle handle);
+
+/*!
+ * \brief Deprecated. Please use \ref TreeliteGTILGetPredictOutputSizeEx instead.
+ */
+TREELITE_DLL int TreeliteGTILGetPredictOutputSize(ModelHandle model, size_t num_row, size_t* out);
+
 /*!
  * \brief Given a batch of data rows, query the necessary size of array to hold predictions for all
  *        data points.
  * \param model Treelite Model object
  * \param num_row Number of rows in the input
+ * \param config Configuration of GTIL predictor. Set this by calling \ref TreeliteGTILParseConfig.
  * \param out Size of output buffer that should be allocated
  * \return 0 for success; -1 for failure
  */
-TREELITE_DLL int TreeliteGTILGetPredictOutputSize(ModelHandle model, size_t num_row, size_t* out);
+TREELITE_DLL int TreeliteGTILGetPredictOutputSizeEx(ModelHandle model, size_t num_row,
+                                                    GTILConfigHandle config, size_t* out);
+
+/*!
+ * \brief Deprecated. Please use \ref TreeliteGTILPredictEx instead.
+ */
+TREELITE_DLL int TreeliteGTILPredict(ModelHandle model, const float* input, size_t num_row,
+                                     float* output, int nthread, int pred_transform,
+                                     size_t* out_result_size);
 
 /*!
  * \brief Predict with a 2D dense array
  * \param model Treelite Model object
  * \param input The 2D data array, laid out in row-major layout
  * \param num_row Number of rows in the data matrix.
- * \param output Pointer to buffer to store the output. Call TreeliteGTILGetPredictOutputSize()
- *               to get the amount of buffer you should allocate for this parameter.
- * \param nthread number of CPU threads to use. Set <= 0 to use all CPU cores.
- * \param pred_transform After computing the prediction score, whether to transform it.
+ * \param output Pointer to buffer to store the output. Call \ref
+ *               TreeliteGTILGetPredictOutputSizeEx to get the amount of buffer you should
+ *               allocate for this parameter.
+ * \param config Configuration of GTIL predictor. Set this by calling \ref TreeliteGTILParseConfig.
  * \param out_result_size Size of output. This could be smaller than
- *                        TreeliteGTILGetPredictOutputSize() but could never be larger than
- *                        TreeliteGTILGetPredictOutputSize().
+ *                        \ref TreeliteGTILGetPredictOutputSizeEx but could never be larger than
+ *                        \ref TreeliteGTILGetPredictOutputSizeEx.
+ * \param out_result_ndim Number of dimensions in the output array.
+ * \param out_result_shape Pointer to an array containing dimensions of the prediction output.
+ *                         This array shall have [out_result_ndim] elements.
+ *                         The product of the elements shall be equal to out_result_size.
  * \return 0 for success; -1 for failure
  */
-TREELITE_DLL int TreeliteGTILPredict(ModelHandle model, const float* input, size_t num_row,
-                                     float* output, int nthread, int pred_transform,
-                                     size_t* out_result_size);
+TREELITE_DLL int TreeliteGTILPredictEx(ModelHandle model, const float* input, size_t num_row,
+                                       float* output, GTILConfigHandle config,
+                                       size_t* out_result_size, size_t* out_result_ndim,
+                                       size_t** out_result_shape);
 
 /*! \} */
 
 /*!
- * \defgroup model_builder
- * Model builder interface: build trees incrementally
+ * \defgroup model_builder Model builder interface: build trees incrementally
  * \{
  */
 /*!
