@@ -9,6 +9,7 @@
 
 #include <treelite/base.h>
 #include <treelite/version.h>
+#include <treelite/logging.h>
 #include <algorithm>
 #include <map>
 #include <memory>
@@ -101,7 +102,7 @@ class ContiguousArray {
 /*!
  * \brief Enum type representing the task type.
  *
- * The task type places constraints on the parameters of TaskParameter. See the docstring for each
+ * The task type places constraints on the parameters of TaskParam. See the docstring for each
  * enum constants for more details.
  */
 enum class TaskType : uint8_t {
@@ -141,8 +142,8 @@ enum class TaskType : uint8_t {
    * are combined via summing or averaging, depending on the value of the [average_tree_output]
    * field. In effect, each tree is casting a set of weighted (fractional) votes for the classes.
    *
-   * An example of kMultiClfProbDistLeaf task type is found in RandomForestClassifier of
-   * scikit-learn.
+   * Examples of kMultiClfProbDistLeaf task type are found in RandomForestClassifier of
+   * scikit-learn and RandomForestClassifier of cuML.
    *
    * The kMultiClfProbDistLeaf task type implies the following constraints on the task parameters:
    * output_type=float, grove_per_class=false, num_class>1, leaf_vector_size=num_class.
@@ -159,8 +160,6 @@ enum class TaskType : uint8_t {
    * Models of type kMultiClfCategLeaf can be converted into the kMultiClfProbDistLeaf type, by
    * converting the output of every leaf node into the equivalent one-hot-encoded vector.
    *
-   * An example of kMultiClfCategLeaf task type is found in RandomForestClassifier of cuML.
-   *
    * The kMultiClfCategLeaf task type implies the following constraints on the task parameters:
    * output_type=int, grove_per_class=false, num_class>1, leaf_vector_size=1.
    */
@@ -169,14 +168,29 @@ enum class TaskType : uint8_t {
 
 inline std::string TaskTypeToString(TaskType type) {
   switch (type) {
-    case TaskType::kBinaryClfRegr: return "BinaryClfRegr";
-    case TaskType::kMultiClfGrovePerClass: return "MultiClfGrovePerClass";
-    case TaskType::kMultiClfProbDistLeaf: return "MultiClfProbDistLeaf";
-    case TaskType::kMultiClfCategLeaf: return "MultiClfCategLeaf";
+    case TaskType::kBinaryClfRegr: return "kBinaryClfRegr";
+    case TaskType::kMultiClfGrovePerClass: return "kMultiClfGrovePerClass";
+    case TaskType::kMultiClfProbDistLeaf: return "kMultiClfProbDistLeaf";
+    case TaskType::kMultiClfCategLeaf: return "kMultiClfCategLeaf";
     default: return "";
   }
 }
 
+inline TaskType StringToTaskType(const std::string& str) {
+  if (str == "kBinaryClfRegr") {
+    return TaskType::kBinaryClfRegr;
+  } else if (str == "kMultiClfGrovePerClass") {
+    return TaskType::kMultiClfGrovePerClass;
+  } else if (str == "kMultiClfProbDistLeaf") {
+    return TaskType::kMultiClfProbDistLeaf;
+  } else if (str == "kMultiClfCategLeaf") {
+    return TaskType::kMultiClfCategLeaf;
+  } else {
+    TREELITE_LOG(FATAL) << "Unknown task type: " << str;
+    return TaskType::kBinaryClfRegr;  // to avoid compiler warning
+  }
+}
+
 /*! \brief Group of parameters that are dependent on the choice of the task type. */
 struct TaskParam {
   enum class OutputType : uint8_t { kFloat = 0, kInt = 1 };
@@ -215,6 +229,17 @@ inline std::string OutputTypeToString(TaskParam::OutputType type) {
   }
 }
 
+inline TaskParam::OutputType StringToOutputType(const std::string& str) {
+  if (str == "float") {
+    return TaskParam::OutputType::kFloat;
+  } else if (str == "int") {
+    return TaskParam::OutputType::kInt;
+  } else {
+    TREELITE_LOG(FATAL) << "Unrecognized output type: " << str;
+    return TaskParam::OutputType::kFloat;  // to avoid compiler warning
+  }
+}
+
 static_assert(std::is_pod<TaskParam>::value, "TaskParameter must be POD type");
 
 /*! \brief in-memory representation of a decision tree */
@@ -662,8 +687,7 @@ class Tree {
 
 struct ModelParam {
   /*!
-  * \defgroup model_param
-  * Extra parameters for tree ensemble models
+  * \defgroup model_param Extra parameters for tree ensemble models
   * \{
   */
   /*!
