@@ -878,14 +878,21 @@ Model::DeserializeTemplate(HeaderPrimitiveFieldHandlerFunc header_primitive_fiel
   header_primitive_field_handler(&minor_ver);
   header_primitive_field_handler(&patch_ver);
   if (major_ver != TREELITE_VER_MAJOR && !(major_ver == 2 && minor_ver == 4)) {
-    std::ostringstream oss;
-    oss << "Cannot deserialize model from a different major Treelite version or "
+    TREELITE_LOG(FATAL)
+        << "Cannot load model from a different major Treelite version or "
         << "a version before 2.4.0." << std::endl
         << "Currently running Treelite version " << TREELITE_VER_MAJOR << "."
         << TREELITE_VER_MINOR << "." << TREELITE_VER_PATCH << std::endl
         << "The model checkpoint was generated from Treelite version " << major_ver << "."
         << minor_ver << "." << patch_ver;
-    throw Error(oss.str());
+  } else if (major_ver == TREELITE_VER_MAJOR && minor_ver > TREELITE_VER_MINOR) {
+    TREELITE_LOG(WARNING)
+        << "The model you are loading originated from a newer Treelite version; some "
+        << "functionalities may be unavailable." << std::endl
+        << "Currently running Treelite version " << TREELITE_VER_MAJOR << "."
+        << TREELITE_VER_MINOR << "." << TREELITE_VER_PATCH << std::endl
+        << "The model checkpoint was generated from Treelite version " << major_ver << "."
+        << minor_ver << "." << patch_ver;
   }
   header_primitive_field_handler(&threshold_type);
   header_primitive_field_handler(&leaf_output_type);
