@@ -17,11 +17,13 @@
 #include <treelite/math.h>
 #include <treelite/gtil.h>
 #include <treelite/logging.h>
+#include <numeric>
 #include <memory>
 #include <algorithm>
 #include <fstream>
 #include <string>
 #include <cstdio>
+#include <cstddef>
 
 using namespace treelite;
 
@@ -31,6 +33,8 @@ namespace {
 struct TreeliteAPIThreadLocalEntry {
   /*! \brief result holder for returning string */
   std::string ret_str;
+  /*! \brief Temp variable for returning prediction shape. */
+  std::vector<std::size_t> prediction_shape;
 };
 
 // define threadlocal store for returning information
@@ -114,6 +118,14 @@ int TreeliteCompilerFree(CompilerHandle handle) {
 }
 
 int TreeliteLoadLightGBMModel(const char* filename, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+      << "TreeliteLoadLightGBMModel() is deprecated. Please use "
+      << "TreeliteLoadLightGBMModelEx() instead.";
+  return TreeliteLoadLightGBMModelEx(filename, "{}", out);
+}
+
+int TreeliteLoadLightGBMModelEx(const char* filename, const char* config_json, ModelHandle* out) {
+  // config_json is unused for now
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadLightGBMModel(filename);
   *out = static_cast<ModelHandle>(model.release());
@@ -121,6 +133,14 @@ int TreeliteLoadLightGBMModel(const char* filename, ModelHandle* out) {
 }
 
 int TreeliteLoadXGBoostModel(const char* filename, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+      << "TreeliteLoadXGBoostModel() is deprecated. Please use "
+      << "TreeliteLoadXGBoostModelEx() instead.";
+  return TreeliteLoadXGBoostModelEx(filename, "{}", out);
+}
+
+int TreeliteLoadXGBoostModelEx(const char* filename, const char* config_json, ModelHandle* out) {
+  // config_json is unused for now
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadXGBoostModel(filename);
   *out = static_cast<ModelHandle>(model.release());
@@ -128,20 +148,45 @@ int TreeliteLoadXGBoostModel(const char* filename, ModelHandle* out) {
 }
 
 int TreeliteLoadXGBoostJSON(const char* filename, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+      << "TreeliteLoadXGBoostJSON() is deprecated. Please use "
+      << "TreeliteLoadXGBoostJSONEx() instead.";
+  return TreeliteLoadXGBoostJSONEx(filename, "{}", out);
+}
+
+int TreeliteLoadXGBoostJSONEx(const char* filename, const char* config_json, ModelHandle* out) {
   API_BEGIN();
-  std::unique_ptr<Model> model = frontend::LoadXGBoostJSONModel(filename);
+  std::unique_ptr<Model> model = frontend::LoadXGBoostJSONModel(filename, config_json);
   *out = static_cast<ModelHandle>(model.release());
   API_END();
 }
 
 int TreeliteLoadXGBoostJSONString(const char* json_str, size_t length, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+    << "TreeliteLoadXGBoostJSONString() is deprecated. Please use "
+    << "TreeliteLoadXGBoostJSONStringEx() instead.";
+  return TreeliteLoadXGBoostJSONStringEx(json_str, length, "{}", out);
+}
+
+int TreeliteLoadXGBoostJSONStringEx(
+    const char* json_str, size_t length, const char* config_json, ModelHandle* out) {
   API_BEGIN();
-  std::unique_ptr<Model> model = frontend::LoadXGBoostJSONModelString(json_str, length);
+  std::unique_ptr<Model> model =
+      frontend::LoadXGBoostJSONModelString(json_str, length, config_json);
   *out = static_cast<ModelHandle>(model.release());
   API_END();
 }
 
 int TreeliteLoadXGBoostModelFromMemoryBuffer(const void* buf, size_t len, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+      << "TreeliteLoadXGBoostModelFromMemoryBuffer() is deprecated. Please use "
+      << "TreeliteLoadXGBoostModelFromMemoryBufferEx() instead.";
+  return TreeliteLoadXGBoostModelFromMemoryBufferEx(buf, len, "{}", out);
+}
+
+int TreeliteLoadXGBoostModelFromMemoryBufferEx(
+    const void* buf, size_t len, const char* config_json, ModelHandle* out) {
+  // config_json is unused for now
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadXGBoostModel(buf, len);
   *out = static_cast<ModelHandle>(model.release());
@@ -149,12 +194,29 @@ int TreeliteLoadXGBoostModelFromMemoryBuffer(const void* buf, size_t len, ModelH
 }
 
 int TreeliteLoadLightGBMModelFromString(const char* model_str, ModelHandle* out) {
+  TREELITE_LOG(WARNING)
+      << "TreeliteLoadLightGBMModelFromString() is deprecated. Please use "
+      << "TreeliteLoadLightGBMModelFromStringEx() instead.";
+  return TreeliteLoadLightGBMModelFromStringEx(model_str, "{}", out);
+}
+
+int TreeliteLoadLightGBMModelFromStringEx(
+    const char* model_str, const char* config_json, ModelHandle* out) {
+  // config_json is unused for now
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadLightGBMModelFromString(model_str);
   *out = static_cast<ModelHandle>(model.release());
   API_END();
 }
 
+int TreeliteBuildModelFromJSONString(
+    const char* json_str, const char* config_json, ModelHandle* out) {
+  API_BEGIN();
+  std::unique_ptr<Model> model = frontend::BuildModelFromJSONString(json_str, config_json);
+  *out = static_cast<ModelHandle>(model.release());
+  API_END();
+}
+
 int TreeliteLoadSKLearnRandomForestRegressor(
     int n_estimators, int n_features, const int64_t* node_count, const int64_t** children_left,
     const int64_t** children_right, const int64_t** feature, const double** threshold,
@@ -195,31 +257,58 @@ int TreeliteLoadSKLearnRandomForestClassifier(
 }
 
 int TreeliteLoadSKLearnGradientBoostingRegressor(
-    int n_estimators, int n_features, const int64_t* node_count, const int64_t** children_left,
+    int n_iter, int n_features, const int64_t* node_count, const int64_t** children_left,
     const int64_t** children_right, const int64_t** feature, const double** threshold,
     const double** value, const int64_t** n_node_samples, const double** weighted_n_node_samples,
     const double** impurity, ModelHandle* out) {
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadSKLearnGradientBoostingRegressor(
-      n_estimators, n_features, node_count, children_left, children_right, feature, threshold,
+      n_iter, n_features, node_count, children_left, children_right, feature, threshold,
       value, n_node_samples, weighted_n_node_samples, impurity);
   *out = static_cast<ModelHandle>(model.release());
   API_END();
 }
 
 int TreeliteLoadSKLearnGradientBoostingClassifier(
-    int n_estimators, int n_features, int n_classes, const int64_t* node_count,
+    int n_iter, int n_features, int n_classes, const int64_t* node_count,
     const int64_t** children_left, const int64_t** children_right, const int64_t** feature,
     const double** threshold, const double** value, const int64_t** n_node_samples,
     const double** weighted_n_node_samples, const double** impurity, ModelHandle* out) {
   API_BEGIN();
   std::unique_ptr<Model> model = frontend::LoadSKLearnGradientBoostingClassifier(
-      n_estimators, n_features, n_classes, node_count, children_left, children_right, feature,
+      n_iter, n_features, n_classes, node_count, children_left, children_right, feature,
       threshold, value, n_node_samples, weighted_n_node_samples, impurity);
   *out = static_cast<ModelHandle>(model.release());
   API_END();
 }
 
+int TreeliteLoadSKLearnHistGradientBoostingRegressor(
+    int n_iter, int n_features, const int64_t* node_count, const int64_t** children_left,
+    const int64_t** children_right, const int64_t** feature, const double** threshold,
+    const int8_t** default_left, const double** value, const int64_t** n_node_samples,
+    const double** gain, const double* baseline_prediction, ModelHandle* out) {
+  API_BEGIN();
+  std::unique_ptr<Model> model = frontend::LoadSKLearnHistGradientBoostingRegressor(
+      n_iter, n_features, node_count, children_left, children_right, feature,
+      threshold, default_left, value, n_node_samples, gain, baseline_prediction);
+  *out = static_cast<ModelHandle>(model.release());
+  API_END();
+}
+
+int TreeliteLoadSKLearnHistGradientBoostingClassifier(
+    int n_iter, int n_features, int n_classes, const int64_t* node_count,
+    const int64_t** children_left, const int64_t** children_right, const int64_t** feature,
+    const double** threshold, const int8_t** default_left, const double** value,
+    const int64_t** n_node_samples, const double** gain, const double* baseline_prediction,
+    ModelHandle* out) {
+  API_BEGIN();
+  std::unique_ptr<Model> model = frontend::LoadSKLearnHistGradientBoostingClassifier(
+      n_iter, n_features, n_classes, node_count, children_left, children_right, feature,
+      threshold, default_left, value, n_node_samples, gain, baseline_prediction);
+  *out = static_cast<ModelHandle>(model.release());
+  API_END();
+}
+
 int TreeliteSerializeModel(const char* filename, ModelHandle handle) {
   API_BEGIN();
   FILE* fp = std::fopen(filename, "wb");
@@ -266,19 +355,68 @@ int TreeliteFreeModel(ModelHandle handle) {
   API_END();
 }
 
+int TreeliteGTILParseConfig(const char* config_json, GTILConfigHandle* out) {
+  API_BEGIN();
+  auto parsed_config = std::make_unique<gtil::Configuration>(config_json);
+  *out = static_cast<GTILConfigHandle>(parsed_config.release());
+  API_END();
+}
+
+int TreeliteGTILDeleteConfig(GTILConfigHandle handle) {
+  API_BEGIN();
+  delete static_cast<gtil::Configuration*>(handle);
+  API_END();
+}
+
 int TreeliteGTILGetPredictOutputSize(ModelHandle model, size_t num_row, size_t* out) {
   API_BEGIN();
+  TREELITE_LOG(WARNING)
+    << "TreeliteGTILGetPredictOutputSize() is deprecated; "
+    << "please use TreeliteGTILGetPredictOutputSizeEx() instead";
   const auto* model_ = static_cast<const Model*>(model);
-  *out = gtil::GetPredictOutputSize(model_, num_row);
+  *out = gtil::GetPredictOutputSize(model_, num_row, gtil::Configuration{});
+  API_END();
+}
+
+int TreeliteGTILGetPredictOutputSizeEx(ModelHandle model, size_t num_row, GTILConfigHandle config,
+                                       size_t* out) {
+  API_BEGIN();
+  const auto* model_ = static_cast<const Model*>(model);
+  const auto* config_ = static_cast<const gtil::Configuration*>(config);
+  *out = gtil::GetPredictOutputSize(model_, num_row, *config_);
   API_END();
 }
 
 int TreeliteGTILPredict(ModelHandle model, const float* input, size_t num_row, float* output,
                         int nthread, int pred_transform, size_t* out_result_size) {
   API_BEGIN();
+  TREELITE_LOG(WARNING)
+    << "TreeliteGTILPredict() is deprecated; please use TreeliteGTILPredictEx() instead.";
   const auto* model_ = static_cast<const Model*>(model);
-  *out_result_size =
-      gtil::Predict(model_, input, num_row, output, nthread, (pred_transform == 1));
+  gtil::Configuration config;
+  config.pred_type = (pred_transform == 1 ? treelite::gtil::PredictType::kPredictDefault
+                                          : treelite::gtil::PredictType::kPredictRaw);
+  config.nthread = nthread;
+  auto& pred_shape = TreeliteAPIThreadLocalStore::Get()->prediction_shape;
+  *out_result_size = gtil::Predict(model_, input, num_row, output, config,
+                                   pred_shape);
+  API_END();
+}
+
+int TreeliteGTILPredictEx(ModelHandle model, const float* input, size_t num_row, float* output,
+                          GTILConfigHandle config, size_t* out_result_size,
+                          size_t* out_result_ndim, size_t** out_result_shape) {
+  API_BEGIN();
+  const auto* model_ = static_cast<const Model*>(model);
+  const auto* config_ = static_cast<const gtil::Configuration*>(config);
+  auto& pred_shape = TreeliteAPIThreadLocalStore::Get()->prediction_shape;
+  *out_result_size = gtil::Predict(model_, input, num_row, output, *config_,
+                                   pred_shape);
+  auto prod = std::accumulate<>(std::begin(pred_shape), std::end(pred_shape), 1,
+                                std::multiplies<>{});
+  TREELITE_CHECK_EQ(prod, *out_result_size);
+  *out_result_ndim = pred_shape.size();
+  *out_result_shape = pred_shape.data();
   API_END();
 }
 
