@@ -1,5 +1,5 @@
 /*!
- * Copyright (c) 2017-2020 by Contributors
+ * Copyright (c) 2017-2021 by Contributors
  * \file ast.h
  * \brief Definition for AST classes
  * \author Hyunsu Cho
@@ -7,18 +7,14 @@
 #ifndef TREELITE_COMPILER_AST_AST_H_
 #define TREELITE_COMPILER_AST_AST_H_
 
-#include <dmlc/optional.h>
+#include <treelite/optional.h>
 #include <treelite/base.h>
 #include <fmt/format.h>
 #include <limits>
 #include <string>
 #include <vector>
 #include <utility>
-
-// forward declaration
-namespace treelite_ast_protobuf {
-class ASTNode;
-}  // namespace treelite_ast_protobuf
+#include <cstdint>
 
 namespace treelite {
 namespace compiler {
@@ -29,8 +25,8 @@ class ASTNode {
   std::vector<ASTNode*> children;
   int node_id;
   int tree_id;
-  dmlc::optional<size_t> data_count;
-  dmlc::optional<double> sum_hess;
+  optional<uint64_t> data_count;
+  optional<double> sum_hess;
   virtual std::string GetDump() const = 0;
   virtual ~ASTNode() = 0;  // force ASTNode to be abstract class
  protected:
@@ -65,13 +61,14 @@ class TranslationUnitNode : public ASTNode {
   }
 };
 
+template <typename ThresholdType>
 class QuantizerNode : public ASTNode {
  public:
-  explicit QuantizerNode(const std::vector<std::vector<tl_float>>& cut_pts)
+  explicit QuantizerNode(const std::vector<std::vector<ThresholdType>>& cut_pts)
     : cut_pts(cut_pts) {}
-  explicit QuantizerNode(std::vector<std::vector<tl_float>>&& cut_pts)
+  explicit QuantizerNode(std::vector<std::vector<ThresholdType>>&& cut_pts)
     : cut_pts(std::move(cut_pts)) {}
-  std::vector<std::vector<tl_float>> cut_pts;
+  std::vector<std::vector<ThresholdType>> cut_pts;
 
   std::string GetDump() const override {
     std::ostringstream oss;
@@ -110,12 +107,12 @@ class ConditionNode : public ASTNode {
     : split_index(split_index), default_left(default_left) {}
   unsigned split_index;
   bool default_left;
-  dmlc::optional<double> gain;
+  optional<double> gain;
 
   std::string GetDump() const override {
     if (gain) {
       return fmt::format("ConditionNode {{ split_index: {}, default_left: {}, gain: {} }}",
-                         split_index, default_left, gain.value());
+                         split_index, default_left, *gain);
     } else {
       return fmt::format("ConditionNode {{ split_index: {}, default_left: {} }}",
                          split_index, default_left);
@@ -123,65 +120,71 @@ class ConditionNode : public ASTNode {
   }
 };
 
+template <typename ThresholdType>
 union ThresholdVariant {
-  tl_float float_val;
+  ThresholdType float_val;
   int int_val;
-  ThresholdVariant(tl_float val) : float_val(val) {}
-  ThresholdVariant(int val) : int_val(val) {}
+  explicit ThresholdVariant(ThresholdType val) : float_val(val) {}
+  explicit ThresholdVariant(int val) : int_val(val) {}
 };
 
+template <typename ThresholdType>
 class NumericalConditionNode : public ConditionNode {
  public:
   NumericalConditionNode(unsigned split_index, bool default_left,
                          bool quantized, Operator op,
-                         ThresholdVariant threshold)
+                         ThresholdVariant<ThresholdType> threshold)
     : ConditionNode(split_index, default_left),
-      quantized(quantized), op(op), threshold(threshold) {}
+      quantized(quantized), op(op), threshold(threshold), zero_quantized(-1) {}
   bool quantized;
   Operator op;
-  ThresholdVariant threshold;
+  ThresholdVariant<ThresholdType> threshold;
+  int zero_quantized;  // quantized value of 0.0f (useful when convert_missing_to_zero is set)
 
   std::string GetDump() const override {
-    return fmt::format("NumericalConditionNode {{ {}, quantized: {}, op: {}, threshold: {} }}",
+    return fmt::format("NumericalConditionNode {{ {}, quantized: {}, op: {}, threshold: {}, "
+                       "zero_quantized: {} }}",
                        ConditionNode::GetDump(), quantized, OpName(op),
-                       (quantized ? fmt::format("{:d}", threshold.int_val)
-                                  : fmt::format("{:f}", threshold.float_val)));
+                       (quantized ? fmt::format("{}", threshold.int_val)
+                                  : fmt::format("{}", threshold.float_val)),
+                       zero_quantized);
   }
 };
 
 class CategoricalConditionNode : public ConditionNode {
  public:
   CategoricalConditionNode(unsigned split_index, bool default_left,
-                           const std::vector<uint32_t>& left_categories,
-                           bool convert_missing_to_zero)
+                           const std::vector<uint32_t>& matching_categories,
+                           bool categories_list_right_child)
     : ConditionNode(split_index, default_left),
-      left_categories(left_categories),
-      convert_missing_to_zero(convert_missing_to_zero) {}
-  std::vector<uint32_t> left_categories;
-  bool convert_missing_to_zero;
+      matching_categories(matching_categories),
+      categories_list_right_child(categories_list_right_child) {}
+  std::vector<uint32_t> matching_categories;
+  bool categories_list_right_child;
 
   std::string GetDump() const override {
     std::ostringstream oss;
     oss << "[";
-    for (const auto& e : left_categories) {
+    for (const auto& e : matching_categories) {
       oss << e << ", ";
     }
     oss << "]";
-    return fmt::format("CategoricalConditionNode {{ {}, left_categories: {}, "
-                       "convert_missing_to_zero: {} }}",
-                       ConditionNode::GetDump(), oss.str(), convert_missing_to_zero);
+    return fmt::format("CategoricalConditionNode {{ {}, matching_categories: {}, "
+                       "categories_list_right_child: {} }}",
+                       ConditionNode::GetDump(), oss.str(), categories_list_right_child);
   }
 };
 
+template <typename LeafOutputType>
 class OutputNode : public ASTNode {
  public:
-  explicit OutputNode(tl_float scalar)
+  explicit OutputNode(LeafOutputType scalar)
     : is_vector(false), scalar(scalar) {}
-  explicit OutputNode(const std::vector<tl_float>& vector)
+  explicit OutputNode(const std::vector<LeafOutputType>& vector)
     : is_vector(true), vector(vector) {}
   bool is_vector;
-  tl_float scalar;
-  std::vector<tl_float> vector;
+  LeafOutputType scalar;
+  std::vector<LeafOutputType> vector;
 
   std::string GetDump() const override {
     if (is_vector) {
